#!/usr/bin/env python
"""
The changedetect module provides tools for change detection using remotely sensed imagery.
"""

# import the C++ extension into this level (when/if implemented)
# from ._changedetect import *
