// the configured options and settings for RSGISLib

#define RSGISLIB_HG_COMMIT 

#define RSGISLIB_VERSION_MAJOR 5
#define RSGISLIB_VERSION_MINOR 0
#define RSGISLIB_VERSION_PATCH 8

#define RSGISLIB_VERSION "5.0.8"
#define RSGISLIB_PACKAGE_VERSION "5.0.8"
#define RSGISLIB_PACKAGE_STRING "RSGISLib 5.0.8"
#define RSGISLIB_PACKAGE_BUGREPORT "rsgislib-support@googlegroups.com"
#define RSGISLIB_PACKAGE "RSGISLib"
#define RSGISLIB_COPYRIGHT_YEAR "2021"

#define RSGISLIB_CALIBRATION_VERSION_MAJOR 1
#define RSGISLIB_CALIBRATION_VERSION_MINOR 1
#define RSGISLIB_CALIBRATION_VERSION_PATCH 0
#define RSGISLIB_CALIBRATION_VERSION "1.1.0"

#define RSGISLIB_CLASSIFY_VERSION_MAJOR 1
#define RSGISLIB_CLASSIFY_VERSION_MINOR 1
#define RSGISLIB_CLASSIFY_VERSION_PATCH 0
#define RSGISLIB_CLASSIFY_VERSION "1.1.0"

#define RSGISLIB_COMMONS_VERSION_MAJOR 1
#define RSGISLIB_COMMONS_VERSION_MINOR 3
#define RSGISLIB_COMMONS_VERSION_PATCH 0
#define RSGISLIB_COMMONS_VERSION "1.3.0"

#define RSGISLIB_DATASTRUCT_VERSION_MAJOR 1
#define RSGISLIB_DATASTRUCT_VERSION_MINOR 1
#define RSGISLIB_DATASTRUCT_VERSION_PATCH 0
#define RSGISLIB_DATASTRUCT_VERSION "1.1.0"

#define RSGISLIB_FILTERING_VERSION_MAJOR 1
#define RSGISLIB_FILTERING_VERSION_MINOR 0
#define RSGISLIB_FILTERING_VERSION_PATCH 0
#define RSGISLIB_FILTERING_VERSION "1.0.0"

#define RSGISLIB_GEOM_VERSION_MAJOR 
#define RSGISLIB_GEOM_VERSION_MINOR 
#define RSGISLIB_GEOM_VERSION_PATCH 
#define RSGISLIB_GEOM_VERSION ""

#define RSGISLIB_IMG_VERSION_MAJOR 3
#define RSGISLIB_IMG_VERSION_MINOR 0
#define RSGISLIB_IMG_VERSION_PATCH 0
#define RSGISLIB_IMG_VERSION "3.0.0"

#define RSGISLIB_MATHS_VERSION_MAJOR 1
#define RSGISLIB_MATHS_VERSION_MINOR 1
#define RSGISLIB_MATHS_VERSION_PATCH 0
#define RSGISLIB_MATHS_VERSION "1.1.0"

#define RSGISLIB_MODELING_VERSION_MAJOR 
#define RSGISLIB_MODELING_VERSION_MINOR 
#define RSGISLIB_MODELING_VERSION_PATCH 
#define RSGISLIB_MODELING_VERSION ""

#define RSGISLIB_RADAR_VERSION_MAJOR 
#define RSGISLIB_RADAR_VERSION_MINOR 
#define RSGISLIB_RADAR_VERSION_PATCH 
#define RSGISLIB_RADAR_VERSION ""

#define RSGISLIB_RASTERGIS_VERSION_MAJOR 2
#define RSGISLIB_RASTERGIS_VERSION_MINOR 0
#define RSGISLIB_RASTERGIS_VERSION_PATCH 0
#define RSGISLIB_RASTERGIS_VERSION "2.0.0"

#define RSGISLIB_REGISTRATION_VERSION_MAJOR 1
#define RSGISLIB_REGISTRATION_VERSION_MINOR 2
#define RSGISLIB_REGISTRATION_VERSION_PATCH 0
#define RSGISLIB_REGISTRATION_VERSION "1.2.0"

#define RSGISLIB_SEGMENTATION_VERSION_MAJOR 1
#define RSGISLIB_SEGMENTATION_VERSION_MINOR 0
#define RSGISLIB_SEGMENTATION_VERSION_PATCH 0
#define RSGISLIB_SEGMENTATION_VERSION "1.0.0"

#define RSGISLIB_UTILS_VERSION_MAJOR 1
#define RSGISLIB_UTILS_VERSION_MINOR 1
#define RSGISLIB_UTILS_VERSION_PATCH 0
#define RSGISLIB_UTILS_VERSION "1.1.0"

#define RSGISLIB_VECTOR_VERSION_MAJOR 2
#define RSGISLIB_VECTOR_VERSION_MINOR 2
#define RSGISLIB_VECTOR_VERSION_PATCH 0
#define RSGISLIB_VECTOR_VERSION "2.2.0"
